/********************************************
 * Character display using PIC18F14K50      *
 * Constructed by Katsumi                   *
 * (kmorimatsu@users.sourceforge.jp)        *
 * This product can be re-distributed under *
 * the LGPL license v2.1.                   *
 * No warranty.                             *
 ********************************************/

#include <p18f14k50.h>


/* For assembler */
#define W 0
#define F 1
#define ACCESS 0
#define BANKED 1

/* Configurations follow */
#pragma config CPUDIV = NOCLKDIV ,USBDIV = OFF, PCLKEN = ON 
#pragma config FOSC = HS, PLLEN = ON, HFOFST = OFF
#pragma config PWRTEN = ON, BOREN = OFF, MCLRE = OFF, BORV = 30
#pragma config WDTEN = OFF, LVP = OFF, FCMEN = OFF, IESO = OFF
#pragma config CP0 = OFF, XINST = OFF

/* video and Hsync lines */
#define sync LATBbits.LATB4
#define video LATCbits.LATC7
#define busy LATBbits.LATB5 // busy = /ready

/* global variables */
#pragma udata g_vram080=0x80
unsigned char g_vram080[0x80];
#pragma udata g_vram100=0x100
unsigned char g_vram100[0x100];
#pragma udata g_vram200=0x200
unsigned char g_vram200[0x100];
#pragma udata access accessram
unsigned near char g_cline;
unsigned near char g_fsrl;
unsigned near char g_fsrh;
unsigned near char g_fsrl_line0;
unsigned near char g_fsrh_line0;
unsigned near char g_ascii;
unsigned near char dummy,dummy1,dummy2,dummy3;
unsigned near int dummy16;
unsigned near char g_PORTB, g_PORTC, g_data_exists;
unsigned near char g_cursor_x,g_cursor_y;
#pragma udata

/* CG-ROM emmbed in program memory */ 
#pragma romdata cgrom=0x3000
#include "cgrom.h"

/* prototyping */
void hpInt(void);
void insertASCII(void);
void blanc(void);
void show(void);
void equalize(void);
void preEqualize(void);
void main(void);
void blink(void);
void clearScreen(void);
void scroll(void);

/* Macros as the inline functions */
#define waitTMR0H(x) do { _asm MOVF TMR0L,W,ACCESS _endasm } while(TMR0H<x)
#define waitTMR0L(x) do {} while(TMR0L<x);

/* interrupt */
#pragma code hpInt=0x08
void hpInt(){
	g_PORTC=PORTC;
	g_PORTB=PORTB;
	INTCONbits.INT0IF=0;
	g_data_exists=1;
	busy=1;
	_asm RETFIE 1 _endasm
}

#pragma code
void test(){
	// Ignore the codes in this function, please.
	// These are used to see the assembler codes corresponding to C codes.
	do {} while (TMR0L & 3); 
	while (TMR0L & 1) { }
	if (busy) {_asm NOP NOP _endasm}
}
void insertASCII(){
	g_data_exists=0;
	/* ASCII code handling routine follows */
	dummy=g_ascii; // "dummy" is the previous code in this function.
	g_ascii=((g_PORTC>>1)&0x3f) | g_PORTB & 0xc0;
	switch(g_ascii){
		case 0x16:
			// Clear screen
			clearScreen();
			g_cursor_x=0;
			g_cursor_y=0;
			return;
		case 0x0a:
			if (dummy==0x0d) return; // "\r\n" support
		case 0x0d:
			// return
			g_cursor_x=0;
			g_cursor_y++;
			if (g_cursor_y==20) {
				scroll();
				g_cursor_y=19;
			}
			return;
		case 0x08:
			// back space
			if (g_cursor_x==0 && g_cursor_y==0) return;
			if ((--g_cursor_x)==255) {
				g_cursor_x=31;
				g_cursor_y--;
			}
			break;
		case 0x13:
			// cursor right
			if (31<(++g_cursor_x)) g_cursor_x=0;
			else return;
		case 0x11:
			// cursor down
			if (19<(++g_cursor_y)) g_cursor_y=0;
			return;
		case 0x14:
			// cursor left
			if ((--g_cursor_x)==255) g_cursor_x=31;
			else return;
		case 0x12:
			// cursor up
			if ((--g_cursor_y)==255) g_cursor_y=19;
			return;
		default:
			break;
	}
	dummy16=g_fsrh_line0;
	dummy16<<=3;
	dummy16+=g_cursor_y;
	dummy16<<=5;
	dummy16+=g_fsrl_line0;
	dummy16+=g_cursor_x;
	if (0x2ff<dummy16) dummy16-=0x280;
	FSR0L=dummy16 & 0xff;
	FSR0H=dummy16>>8;
	if (g_ascii==0x08) {
		// back space
		INDF0=0x20;
		return;
	}
	INDF0=g_ascii;
	if (31<(++g_cursor_x)) {
		g_cursor_x=0;
		g_cursor_y++;
		if (g_cursor_y==20) {
			scroll();
			g_cursor_y=19;
		}
	}
}
void blanc(){
	do {} while(INTCONbits.TMR0IF==0);
	TMR0H=0xFD;
	TMR0L+=0x05;
	INTCONbits.TMR0IF=0;
	// 0.3 V until 1.5 usec	
	sync=1;
	video=0;
	do {} while (TMR0L & 2); 
	do {} while (TMR0L & 1); 
	waitTMR0L(0x1D);
	// 0 V until 6.2 usec
	sync=0;
	waitTMR0L(0x55);
	// 0.3 V until end of raster
	sync=1;
}


void show(){
	do {} while(INTCONbits.TMR0IF==0);
	TMR0H=0xFD;
	TMR0L+=0x05;
	INTCONbits.TMR0IF=0;
	// 0.3 V until 1.5 usec	
	sync=1;
	video=0;
	do {} while (TMR0L & 2); 
	do {} while (TMR0L & 1); 
	waitTMR0L(0x1D);
	// 0 V until 6.2 usec
	sync=0;
	waitTMR0L(0x55);
	// 0.3 V until 15.9 usec
	sync=1;

	// Register preparation for showing
	TBLPTRU=0;
	TBLPTRH=0x30|g_cline;
	FSR0H=g_fsrh;
	FSR0L=g_fsrl;

	// Wait until 15.6 usec (10 cycles earlier)
	waitTMR0L(0xB6);

	// Main showing routine follows
_asm

MOVF POSTINC0, W, ACCESS       // 0 Copy the value in RAM to W, then increment the pointer
MOVWF TBLPTRL, ACCESS          // 0 Copy the value from W to TBLPTRL
TBLRD                          // 0 Read the CGROM data to TABLAT
MOVF POSTINC0, W, ACCESS       // 1 Copy the value in RAM to W, then increment the pointer
MOVWF TBLPTRL, ACCESS          // 1 Copy the value from W to TBLPTRL
MOVF TABLAT, W, ACCESS         // 0 Copy the value from TABLAT to W
TBLRD                          // 1 Read the CGROM data to TABLAT
// column 0
MOVWF LATC, ACCESS    // bit 7 // 0 Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // 2 Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // 2 Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // 1 Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // 2 Read the CGROM data to TABLAT
// column 1
MOVWF LATC, ACCESS    // bit 7 // 1 Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // 2 Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 2
MOVWF LATC, ACCESS    // bit 7 // 2 Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 3
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 4
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 5
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 6
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 7
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 8
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 9
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 10
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 11
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 12
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 13
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 14
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 15
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 16
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 17
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 18
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 19
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 20
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 21
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 22
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 23
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 24
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 25
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 26
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 27
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 28
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 29
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
MOVF POSTINC0, W, ACCESS       // Copy the value in RAM to W, then increment the pointer
RLNCF LATC, F, ACCESS // bit 2
MOVWF TBLPTRL, ACCESS          // Copy the value from W to TBLPTRL
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
TBLRD                          // Read the CGROM data to TABLAT
// column 30
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
NOP
RLNCF LATC, F, ACCESS // bit 2
NOP
RLNCF LATC, F, ACCESS // bit 1
MOVF TABLAT, W, ACCESS         // Copy the value from TABLAT to W
RLNCF LATC, F, ACCESS // bit 0
NOP
NOP
// column 31
MOVWF LATC, ACCESS    // bit 7 // Copy the value from W to LATC
NOP
RLNCF LATC, F, ACCESS // bit 6
NOP
RLNCF LATC, F, ACCESS // bit 5
NOP
RLNCF LATC, F, ACCESS // bit 4
NOP
RLNCF LATC, F, ACCESS // bit 3
NOP
RLNCF LATC, F, ACCESS // bit 2
NOP
RLNCF LATC, F, ACCESS // bit 1
NOP
RLNCF LATC, F, ACCESS // bit 0
NOP
NOP
_endasm
	video=0;
}


void equalize(){
	do {} while(INTCONbits.TMR0IF==0);
	TMR0H=0xFD;
	TMR0L+=0x05;
	INTCONbits.TMR0IF=0;
	// 0 V until 27.08 usec
	sync=0;
	video=0;
	waitTMR0H(0xFE);
	waitTMR0L(0x50);
	// 0.3 V until 31.78
	sync=1;
	while(TMR0L<0x88);
	// 0 V until 58.76 usec
	sync=0;
	waitTMR0H(0xFF);
	waitTMR0L(0xCC);
	// 0.3 v until end of raster
	sync=1;
}
void preEqualize(){
	do {} while(INTCONbits.TMR0IF==0);
	TMR0H=0xFD;
	TMR0L+=0x05;
	INTCONbits.TMR0IF=0;
	// 0 V until 2.3 usec
	sync=0;
	video=0;
	waitTMR0L(0x28);
	// 0.3 V until 31.78 usec
	sync=1;
	waitTMR0H(0xFE);
	waitTMR0L(0x88);
	// 0 v until 34.08 usec
	sync=0;
	waitTMR0L(0xa4);
	// 0.3 V until end of raster
	sync=1;
}
#define postEqualize() preEqualize()

/*
RA4: crystal
RA5: crystal
RC0: INT0, /WR
RC1: D0
RC2: D1
RC3: D2
RC4: D3
RC5: D4
RC6: D5
RC7: video
RB4: /Hsync
RB5: /ready (busy)
RB6: D6
RB7: D7
*/

void main(){
	unsigned char i;
	/* initialization */
	// Timer0 on
	T0CON=0x88;
	// PORT A: all input
	TRISA=0xFF;
	// PORT B; RB4, RB5: output, RB6, RB7: input
	ANSELH=0x00;
	LATB=0x30;LATB=0x00;
	TRISB=0xC0;
	// PORT C; RC7: output, others: input
	ANSEL=0x00;
	LATC=0x00;
	TRISC=0x7F;
	// Interrupt settings
	RCONbits.IPEN=0;
	INTCON=0x90; // GIE=1, PEIE=0, TMR0IE=0, INT0IE=1, RBIE=0
	INTCON=0xD0; // GIE=1, PEIE=1, TMR0IE=0, INT0IE=1, RBIE=0
	INTCON2bits.INTEDG0=0; // interrupt when signal falles to L

	// initialize
	g_fsrl_line0=0x80;
	g_fsrh_line0=0x00;
	busy=0;
	g_cursor_x=0;
	g_cursor_y=0;
	g_data_exists=0;
	g_ascii=0;
	clearScreen();

	i=0;
	do {
		g_vram100[i]=i;
		g_vram200[i]=i;
		if (i<0x80) {
			g_vram080[i]=i+0x40;
		}
	} while (++i!=0);
	while(1){
		preEqualize();
		preEqualize();
		preEqualize();
		equalize();
		equalize();
		equalize();
		postEqualize();
		postEqualize();
		postEqualize();

		i=52;
		do {
			blanc();
			if (g_data_exists) insertASCII();
			else busy=4<i ? 0:1;
		} while(--i);

		g_fsrl=g_fsrl_line0;
		g_fsrh=g_fsrh_line0;
		i=20;
		do {
			if (g_fsrh==0x03) {
				g_fsrl=0x80;
				g_fsrh=0x00;
			}
			g_cline=0;
			do {show();} while ((++g_cline)<8);
			g_fsrh=FSR0H;
			g_fsrl=FSR0L;
		} while(--i);

		i=41;
		do {
			blanc();
			if (g_data_exists) insertASCII();
			else busy=4<i ? 0:1;
		} while(--i);
	}
}
void scroll(){
	FSR0L=g_fsrl_line0;
	FSR0H=g_fsrh_line0;
	_asm
		MOVLW 0x20
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
		MOVWF POSTINC0,ACCESS
	_endasm
	g_fsrl_line0+=0x20;
	if (g_fsrl_line0==0) {
		g_fsrh_line0++;
		if (g_fsrl_line0==0x03) {
			g_fsrl_line0=0x80;
			g_fsrh_line0=0x00;
		}
	}
}
void clearScreen(){
	FSR0L=0x80;
	FSR0H=0x00;
	do {
		_asm
			MOVLW 0x20
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
			MOVWF POSTINC0,ACCESS
		_endasm
	} while (FSR0H<0x03);
}
void blink(){
	// For debugging
	for(dummy2=0;dummy2<10;dummy2++){
		for(dummy=0;dummy<255;dummy++){for(dummy1=0;dummy1<255;dummy1++);}
		busy=1;
		for(dummy=0;dummy<255;dummy++){for(dummy1=0;dummy1<255;dummy1++);}
		busy=0;
	}
}
